>From 95e9c3ceeed63b31ae168b8a7a7a8d0fa8d932bc Mon Sep 17 00:00:00 2001
From: Steinar H. Gunderson <steinar+vlc@gunderson.no>
Date: Sun, 26 Sep 2010 00:24:09 +0200
Subject: [PATCH] Support DVD LPCM encode.

This patch adds encode support to the LPCM module; only the DVD variant for
now. It's not been tested with any other clients than VLC.

The rationale for this functionality is relatively simple -- LPCM is
currently the only low-latency codec that VLC supports and that can be
embedded in a TS stream. Longer-term support for other low-latency codecs
would be good, but if you're not particularly bandwidth-constrained and need
low latency, this is currently the best option I know of.
---
 NEWS                 |    1 +
 modules/codec/lpcm.c |  188 ++++++++++++++++++++++++++++++++++++++++++++++++++
 2 files changed, 189 insertions(+), 0 deletions(-)

diff --git a/NEWS b/NEWS
index f61d59b..dd935d4 100644
--- a/NEWS
+++ b/NEWS
@@ -19,6 +19,7 @@ Access
 Codecs
  * You can now use ffmpeg-mt in conjunction with vlc
  * Important fixes for RealVideo 3.0 and 4.0 playback
+ * Encode support for DVD LPCM
 
 Demuxers:
  * id3tag plugin is removed (superseded by taglib).
diff --git a/modules/codec/lpcm.c b/modules/codec/lpcm.c
index 3928065..1360a4f 100644
--- a/modules/codec/lpcm.c
+++ b/modules/codec/lpcm.c
@@ -9,6 +9,7 @@
  *          Christophe Massiot <massiot@via.ecp.fr>
  *          Gildas Bazin <gbazin@videolan.org>
  *          Lauren Aimar <fenrir _AT_ videolan _DOT_ org >
+ *          Steinar H. Gunderson <steinar+vlc@gunderson.no>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -45,6 +46,12 @@ static int  OpenDecoder   ( vlc_object_t * );
 static int  OpenPacketizer( vlc_object_t * );
 static void CloseCommon   ( vlc_object_t * );
 
+#ifdef ENABLE_SOUT
+static int  OpenEncoder   ( vlc_object_t * );
+static void CloseEncoder  ( vlc_object_t * );
+static block_t *EncodeFrames( encoder_t *, aout_buffer_t * );
+#endif
+
 vlc_module_begin ()
 
     set_category( CAT_INPUT )
@@ -58,6 +65,14 @@ vlc_module_begin ()
     set_capability( "packetizer", 100 )
     set_callbacks( OpenPacketizer, CloseCommon )
 
+#ifdef ENABLE_SOUT
+    add_submodule ()
+    set_description( N_("Linear PCM audio encoder") )
+    set_capability( "encoder", 100 )
+    set_callbacks( OpenEncoder, CloseEncoder )
+    add_shortcut( "lpcm" )
+#endif
+
 vlc_module_end ()
 
 
@@ -79,6 +94,19 @@ struct decoder_sys_t
     int      i_type;
 };
 
+#ifdef ENABLE_SOUT
+struct encoder_sys_t
+{
+    int     i_channels;
+    int     i_rate;
+
+    int     i_frame_samples;
+    uint8_t *p_buffer;
+    int     i_buffer_used;
+    int     i_frame_num;
+};
+#endif
+
 /*
  * LPCM DVD header :
  * - number of frames in this packet (8 bits)
@@ -399,6 +427,166 @@ static void CloseCommon( vlc_object_t *p_this )
     free( p_dec->p_sys );
 }
 
+#ifdef ENABLE_SOUT
+/*****************************************************************************
+ * OpenEncoder: lpcm encoder construction
+ *****************************************************************************/
+static int OpenEncoder( vlc_object_t *p_this )
+{
+    encoder_t *p_enc = (encoder_t *)p_this;
+    encoder_sys_t *p_sys;
+
+    /* We only support DVD LPCM yet. */
+    if( p_enc->fmt_out.i_codec != VLC_CODEC_DVD_LPCM )
+        return VLC_EGENERIC;
+
+    if( p_enc->fmt_in.audio.i_rate != 48000 &&
+        p_enc->fmt_in.audio.i_rate != 96000 &&
+        p_enc->fmt_in.audio.i_rate != 44100 &&
+        p_enc->fmt_in.audio.i_rate != 32000 )
+    {
+        msg_Err( p_enc, "DVD LPCM supports only sample rates of 48, 96, 44.1 or 32 kHz" );
+        return VLC_EGENERIC;
+    }
+
+    if( p_enc->fmt_in.audio.i_channels > 8 )
+    {
+        msg_Err( p_enc, "DVD LPCM supports a maximum of eight channels" );
+        return VLC_EGENERIC;
+    }
+
+    /* Allocate the memory needed to store the encoder's structure */
+    if( ( p_enc->p_sys = p_sys =
+          (encoder_sys_t *)malloc(sizeof(encoder_sys_t)) ) == NULL )
+        return VLC_ENOMEM;
+
+    /* In DVD LCPM, a frame is always 150 PTS ticks. */
+    p_sys->i_frame_samples = p_enc->fmt_in.audio.i_rate * 150 / 90000;
+    p_sys->p_buffer = (uint8_t *)malloc(
+        p_sys->i_frame_samples *
+        p_enc->fmt_in.audio.i_channels *
+        p_enc->fmt_in.audio.i_bitspersample);
+    p_sys->i_buffer_used = 0;
+    p_sys->i_frame_num = 0;
+
+    p_sys->i_channels = p_enc->fmt_in.audio.i_channels;
+    p_sys->i_rate = p_enc->fmt_in.audio.i_rate;
+
+    p_enc->pf_encode_audio = EncodeFrames;
+    p_enc->fmt_in.i_codec = p_enc->fmt_out.i_codec;
+
+    p_enc->fmt_in.audio.i_bitspersample = 16;
+    p_enc->fmt_in.i_codec = VLC_CODEC_S16B;
+
+    p_enc->fmt_out.i_bitrate =
+        p_enc->fmt_in.audio.i_channels *
+        p_enc->fmt_in.audio.i_rate *
+        p_enc->fmt_in.audio.i_bitspersample *
+        (p_sys->i_frame_samples + LPCM_VOB_HEADER_LEN) /
+        p_sys->i_frame_samples;
+
+    return VLC_SUCCESS;
+}
+
+/*****************************************************************************
+ * CloseEncoder: lpcm encoder destruction
+ *****************************************************************************/
+static void CloseEncoder ( vlc_object_t *p_this )
+{
+    encoder_t     *p_enc = (encoder_t *)p_this;
+    encoder_sys_t *p_sys = p_enc->p_sys;
+
+    free( p_sys->p_buffer );
+    free( p_sys );
+}
+
+/*****************************************************************************
+ * EncodeFrames: encode zero or more LCPM audio packets
+ *****************************************************************************/
+static block_t *EncodeFrames( encoder_t *p_enc, aout_buffer_t *p_aout_buf )
+{
+    encoder_sys_t *p_sys = p_enc->p_sys;
+    block_t *p_first_block = NULL, *p_last_block = NULL;
+
+    if( !p_aout_buf || !p_aout_buf->i_buffer ) return NULL;
+
+    const int i_num_frames = ( p_sys->i_buffer_used + p_aout_buf->i_nb_samples ) /
+        p_sys->i_frame_samples;
+    const int i_leftover_samples = ( p_sys->i_buffer_used + p_aout_buf->i_nb_samples ) %
+        p_sys->i_frame_samples;
+    const int i_frame_size = p_sys->i_frame_samples * p_sys->i_channels * 2 + LPCM_VOB_HEADER_LEN;
+    const int i_start_offset = -p_sys->i_buffer_used;
+
+    uint8_t i_freq_code = 0;
+
+    switch( p_sys->i_rate ) {
+    case 48000:
+        i_freq_code = 0;
+        break;
+    case 96000:
+        i_freq_code = 1;
+        break;
+    case 44100:
+        i_freq_code = 2;
+        break;
+    case 32000:
+        i_freq_code = 3;
+        break;
+    default:
+        assert(0);
+    }
+
+    int i_bytes_consumed = 0;
+
+    for ( int i = 0; i < i_num_frames; ++i )
+    {
+        block_t *p_block = block_New( p_enc, i_frame_size );
+        if( !p_block )
+            return NULL;
+
+        uint8_t *frame = (uint8_t *)p_block->p_buffer;
+        frame[0] = 1;  /* one frame in packet */
+        frame[1] = 0;
+        frame[2] = 0;  /* no first access unit */
+        frame[3] = (p_sys->i_frame_num + i) & 0x1f;  /* no emphasis, no mute */
+        frame[4] = (i_freq_code << 4) | (p_sys->i_channels - 1);
+        frame[5] = 0x80;  /* neutral dynamic range */
+
+        const int i_consume_samples = p_sys->i_frame_samples - p_sys->i_buffer_used;
+        const int i_kept_bytes = p_sys->i_buffer_used * p_sys->i_channels * 2;
+        const int i_consume_bytes = i_consume_samples * p_sys->i_channels * 2;
+
+        memcpy( frame + 6, p_sys->p_buffer, i_kept_bytes );
+        memcpy( frame + 6 + i_kept_bytes, p_aout_buf->p_buffer + i_bytes_consumed, i_consume_bytes );
+
+        p_sys->i_frame_num++;
+        p_sys->i_buffer_used = 0;
+        i_bytes_consumed += i_consume_bytes;
+
+        /* We need to find i_length by means of next_pts due to possible roundoff errors. */
+        mtime_t this_pts = p_aout_buf->i_pts +
+            (i * p_sys->i_frame_samples + i_start_offset) * CLOCK_FREQ / p_sys->i_rate;
+        mtime_t next_pts = p_aout_buf->i_pts +
+            ((i + 1) * p_sys->i_frame_samples + i_start_offset) * CLOCK_FREQ / p_sys->i_rate;
+
+        p_block->i_pts = p_block->i_dts = this_pts;
+        p_block->i_length = next_pts - this_pts;
+
+        if( !p_first_block )
+            p_first_block = p_last_block = p_block;
+        else
+            p_last_block = p_last_block->p_next = p_block;
+    }
+
+    memcpy( p_sys->p_buffer,
+            p_aout_buf->p_buffer + i_bytes_consumed,
+            i_leftover_samples * p_sys->i_channels * 2 );
+    p_sys->i_buffer_used = i_leftover_samples;
+
+    return p_first_block;
+}
+#endif
+
 /*****************************************************************************
  *
  *****************************************************************************/
-- 
1.7.1

